@@ -2,75 +2,66 @@ from __future__ import annotations
 
 from gi.repository import Gdk, Gio, GLib, GObject, Gtk
 
-from gaphor import UML
-from gaphor.abc import ActionProvider
+from gaphor.abc import ActionProvider, TreeItem, TreeModel
 from gaphor.action import action
 from gaphor.core import event_handler
-from gaphor.core.modeling import (
-    DerivedAdded,
-    DerivedDeleted,
-    DerivedSet,
-    Diagram,
-    Element,
-    ElementCreated,
-    ElementDeleted,
-    ElementUpdated,
-    ModelFlushed,
-    ModelReady,
-    Relationship,
-)
+from gaphor.core.modeling import Base, Diagram
 from gaphor.diagram.deletable import deletable
-from gaphor.diagram.diagramtoolbox import DiagramType
 from gaphor.diagram.event import DiagramOpened, DiagramSelectionChanged
-from gaphor.diagram.group import change_owner
+from gaphor.diagram.group import Root, RootType, change_owner, owner
 from gaphor.diagram.tools.dnd import ElementDragData
 from gaphor.event import Notification
-from gaphor.i18n import gettext, translated_ui_string
+from gaphor.i18n import gettext
+from gaphor.services.modelinglanguage import ModelingLanguageChanged
 from gaphor.transaction import Transaction
 from gaphor.ui.abc import UIComponent
-from gaphor.ui.actiongroup import create_action_group
+from gaphor.ui.actiongroup import apply_action_group
 from gaphor.ui.event import (
+    ElementFocused,
     ElementOpened,
     ModelSelectionChanged,
 )
-from gaphor.ui.treemodel import (
-    RelationshipItem,
-    TreeItem,
-    TreeModel,
-    tree_item_sort,
-    visible,
-)
 from gaphor.ui.treesearch import search, sorted_tree_walker
 
 START_EDIT_DELAY = 100  # ms
 
 
 class ModelBrowser(UIComponent, ActionProvider):
-    def __init__(self, event_manager, element_factory, modeling_language):
+    def __init__(
+        self, event_manager, component_registry, element_factory, modeling_language
+    ):
         self.event_manager = event_manager
+        self.component_registry = component_registry
         self.element_factory = element_factory
         self.modeling_language = modeling_language
-        self.model = TreeModel()
+        self.model: TreeModel | None = None
         self.search_bar = None
         self._selection_changed_id = 0
+        self.sorter = None
+        self.selection = None
+        self.tree_view = None
+        self.top_level = None
 
     def open(self):
-        self.event_manager.subscribe(self.on_element_created)
-        self.event_manager.subscribe(self.on_element_deleted)
-        self.event_manager.subscribe(self.on_owner_changed)
-        self.event_manager.subscribe(self.on_owned_element_changed)
-        self.event_manager.subscribe(self.on_attribute_changed)
-        self.event_manager.subscribe(self.on_model_ready)
         self.event_manager.subscribe(self.on_diagram_selection_changed)
+        self.event_manager.subscribe(self.on_modeling_language_changed)
+
+        model_type = self.modeling_language.model_browser_model
+        model = self.component_registry.partial(model_type)(
+            on_select=self._on_select, on_sync=self._on_sync
+        )
+        self.model = model
 
         tree_model = Gtk.TreeListModel.new(
-            self.model.root,
+            model.root,
             passthrough=False,
             autoexpand=False,
-            create_func=self.model.child_model,
-            user_data=None,
+            create_func=model.child_model,
         )
 
+        def tree_item_sort(a, b, *user_data):
+            return model.tree_item_sort(a, b)
+
         self.sorter = Gtk.CustomSorter.new(tree_item_sort)
         tree_sorter = Gtk.TreeListRowSorter.new(self.sorter)
         sort_model = Gtk.SortListModel.new(tree_model, tree_sorter)
@@ -83,6 +74,7 @@ class ModelBrowser(UIComponent, ActionProvider):
             self.selection,
             self.event_manager,
             self.modeling_language,
+            model.template,
         )
         self.tree_view = Gtk.ListView.new(self.selection, factory)
         self.tree_view.set_vexpand(True)
@@ -96,8 +88,8 @@ class ModelBrowser(UIComponent, ActionProvider):
         )
 
         def list_view_activate(list_view, position):
-            if element := self.selection.get_item(position).get_item().element:
-                self.open_element(element)
+            if element := self.selection.get_item(position).get_item().element:  # type: ignore[union-attr]
+                self.focus_element(element)
 
         self.tree_view.connect("activate", list_view_activate)
 
@@ -105,9 +97,7 @@ class ModelBrowser(UIComponent, ActionProvider):
         scrolled_window.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
         scrolled_window.set_child(self.tree_view)
 
-        action_group, shortcuts = create_action_group(self, "tree-view")
-        scrolled_window.insert_action_group("tree-view", action_group)
-        self.tree_view.add_controller(create_shortcut_controller(shortcuts))
+        apply_action_group(self, "tree-view", self.tree_view)
 
         self.tree_view.add_controller(
             create_popup_controller(
@@ -122,39 +112,53 @@ class ModelBrowser(UIComponent, ActionProvider):
         box = Gtk.Box.new(Gtk.Orientation.VERTICAL, 0)
         box.append(self.search_bar)
         box.append(scrolled_window)
-
-        self.on_model_ready()
+        self.top_level = box
 
         return box
 
     def close(self):
-        self.event_manager.unsubscribe(self.on_element_created)
-        self.event_manager.unsubscribe(self.on_element_deleted)
-        self.event_manager.unsubscribe(self.on_owner_changed)
-        self.event_manager.unsubscribe(self.on_owned_element_changed)
-        self.event_manager.unsubscribe(self.on_attribute_changed)
-        self.event_manager.unsubscribe(self.on_model_ready)
         self.event_manager.unsubscribe(self.on_diagram_selection_changed)
+        self.event_manager.unsubscribe(self.on_modeling_language_changed)
+        if self.model:
+            self.model.shutdown()
+        self.model = None
+        self.search_bar = None
+        self._selection_changed_id = 0
+        self.sorter = None
+        self.selection = None
+        self.tree_view = None
+        self.top_level = None
 
-    def select_element(self, element: Element) -> int | None:
-        return select_element(self.tree_view, element)
+    def select_element(self, element: Base) -> int | None:
+        assert self.model
+        return select_element(self.model, self.tree_view, element)
 
     def select_element_quietly(self, element):
         """Select element, but do not trigger a ModelSelectionChanged event."""
+        if not self.selection:
+            return
+
         self.selection.handler_block(self._selection_changed_id)
         try:
             self.select_element(element)
         finally:
             self.selection.handler_unblock(self._selection_changed_id)
 
-    def get_selected_elements(self) -> list[Element]:
+    def get_selected_elements(self) -> list[Base]:
         assert self.model
         return get_selected_elements(self.selection)
 
-    def get_selected_element(self) -> Element | None:
+    def get_selected_element(self) -> Base | None:
         assert self.model
         return next(iter(self.get_selected_elements()), None)
 
+    def focus_element(self, element):
+        assert element
+        if isinstance(element, Diagram):
+            self.event_manager.handle(DiagramOpened(element))
+        else:
+            self.event_manager.handle(ElementFocused(element))
+
     def open_element(self, element):
         assert element
         if isinstance(element, Diagram):
@@ -162,10 +166,18 @@ class ModelBrowser(UIComponent, ActionProvider):
         else:
             self.event_manager.handle(ElementOpened(element))
 
+    def _on_select(self, element):
+        self.select_element_quietly(element)
+
+    def _on_sync(self):
+        if self.sorter:
+            self.sorter.changed(Gtk.SorterChange.DIFFERENT)
+
     @action(name="tree-view.open")
     def tree_view_open_selected(self):
         element = self.get_selected_element()
-        self.open_element(element)
+        if element:
+            self.open_element(element)
 
     @action(name="tree-view.show-in-diagram")
     def tree_view_show_in_diagram(self, diagram_id: str) -> None:
@@ -175,25 +187,23 @@ class ModelBrowser(UIComponent, ActionProvider):
     @action(name="tree-view.rename", shortcut="F2")
     def tree_view_rename_selected(self):
         if row_item := get_first_selected_item(self.selection):
-            tree_item: TreeItem = row_item.get_item()
+            tree_item = row_item.get_item()
             GLib.timeout_add(START_EDIT_DELAY, tree_item.start_editing)
 
-    def _diagram_type_or(self, id: str, default_diagram: DiagramType) -> DiagramType:
-        return next(
-            (dt for dt in self.modeling_language.diagram_types if dt.id == id),
-            default_diagram,
-        )
-
     @action(name="win.create-diagram")
     def tree_view_create_diagram(self, diagram_kind: str):
-        element = self.get_selected_element()
+        element: Base | RootType | None = self.get_selected_element()
 
         with Transaction(self.event_manager):
-            diagram_type = self._diagram_type_or(
-                diagram_kind, DiagramType(diagram_kind, "New Diagram", ())
+            diagram_type = next(
+                dt
+                for dt in self.modeling_language.diagram_types
+                if dt.id == diagram_kind
             )
             try:
-                diagram = diagram_type.create(self.element_factory, element)
+                diagram = diagram_type.create(
+                    self.element_factory, None if element is Root else element
+                )
                 diagram.name = diagram.gettext("New Diagram")
             except TypeError as e:
                 self.event_manager.handle(Notification(str(e)))
@@ -211,31 +221,18 @@ class ModelBrowser(UIComponent, ActionProvider):
 
     @action(name="tree-view.create-element")
     def tree_view_create_element(self, id: str):
-        owner = self.get_selected_element()
+        own = self.get_selected_element()
         element_def = self.element_type(id)
 
         with Transaction(self.event_manager):
             element = self.element_factory.create(element_def.element_type)
             element.name = element_def.name
-            if owner:
-                change_owner(owner, element)
+            if own:
+                change_owner(own, element)
 
             self.select_element(element)
             self.tree_view_rename_selected()
 
-    @action(name="tree-view.create-package")
-    def tree_view_create_package(self):
-        element = self.get_selected_element()
-        with Transaction(self.event_manager):
-            package = self.element_factory.create(UML.Package)
-            if isinstance(element, UML.Package):
-                package.package = element
-                package.name = gettext("{name} package").format(name=element.name)
-            else:
-                package.name = gettext("New package")
-        self.select_element(package)
-        self.tree_view_rename_selected()
-
     @action(name="tree-view.delete", shortcut="Delete")
     def tree_view_delete(self):
         with Transaction(self.event_manager):
@@ -253,48 +250,6 @@ class ModelBrowser(UIComponent, ActionProvider):
         if element := self.element_factory.lookup(id):
             self.select_element(element)
 
-    @event_handler(ElementCreated)
-    def on_element_created(self, event: ElementCreated):
-        self.model.add_element(event.element)
-
-    @event_handler(ElementDeleted)
-    def on_element_deleted(self, event: ElementDeleted):
-        self.model.remove_element(event.element)
-
-    @event_handler(DerivedAdded, DerivedDeleted)
-    def on_owned_element_changed(self, event):
-        """Ensure we update the node once owned elements change."""
-        if event.property in (Element.ownedElement, UML.Namespace.member):
-            self.model.notify_child_model(event.element)
-
-    @event_handler(DerivedSet)
-    def on_owner_changed(self, event: DerivedSet):
-        # Should check on ownedElement as well, since it may not have been updated
-        # before this thing triggers
-        if (
-            event.property not in (Element.owner, Element.memberNamespace)
-        ) or not visible(event.element):
-            return
-        element = event.element
-        self.model.remove_element(element, former_owner=event.old_value)
-        self.model.add_element(element)
-        self.select_element_quietly(element)
-
-    @event_handler(ElementUpdated)
-    def on_attribute_changed(self, event: ElementUpdated):
-        self.model.sync(event.element)
-        self.sorter.changed(Gtk.SorterChange.DIFFERENT)
-
-    @event_handler(ModelReady, ModelFlushed)
-    def on_model_ready(self, _event=None):
-        model = self.model
-        model.clear()
-
-        for element in self.element_factory.select(
-            lambda e: (e.owner is None) and visible(e)
-        ):
-            model.add_element(element)
-
     @event_handler(DiagramSelectionChanged)
     def on_diagram_selection_changed(self, event):
         if not event.focused_item:
@@ -302,6 +257,20 @@ class ModelBrowser(UIComponent, ActionProvider):
         if element := event.focused_item.subject:
             self.select_element_quietly(element)
 
+    @event_handler(ModelingLanguageChanged)
+    def on_modeling_language_changed(self, event: ModelingLanguageChanged) -> None:
+        """Reconfigures the model browser based on the modeling language selected."""
+        if (not self.top_level) or type(
+            self.model
+        ) is self.modeling_language.model_browser_model:
+            return
+
+        bin = self.top_level.get_parent()
+        self.close()
+        widget = self.open()
+        if bin:
+            bin.set_child(widget)
+
 
 class SearchEngine:
     def __init__(self, model, tree_view):
@@ -319,7 +288,7 @@ class SearchEngine:
                 from_current=True,
             ),
         ):
-            select_element(self.tree_view, next_item.element)
+            select_element(self.model, self.tree_view, next_item.element)
 
     def search_next(self, search_text):
         selected_item = get_first_selected_item(self.selection)
@@ -331,10 +300,10 @@ class SearchEngine:
                 from_current=False,
             ),
         ):
-            select_element(self.tree_view, next_item.element)
+            select_element(self.model, self.tree_view, next_item.element)
 
 
-def get_selected_elements(selection: Gtk.SelectionModel) -> list[Element]:
+def get_selected_elements(selection: Gtk.SelectionModel) -> list[Base]:
     bitset = selection.get_selection()
     return [
         e
@@ -350,18 +319,18 @@ def get_first_selected_item(selection):
 
 
 def select_element(
-    tree_view: Gtk.ListView, element: Element, unselect_rest=True
+    model: TreeModel, tree_view: Gtk.ListView, element: Base, unselect_rest=True
 ) -> int | None:
     def expand_up_to_element(element, expand=False) -> int | None:
-        if not element:
+        if element in (Root, None):
             return 0
-        if (n := expand_up_to_element(element.owner, expand=True)) is None:
+        if (n := expand_up_to_element(owner(element), expand=True)) is None:
             return None
-        is_relationship = isinstance(element, Relationship)
         while row := selection.get_item(n):
-            if is_relationship and isinstance(row.get_item(), RelationshipItem):
+            item: TreeItem = row.get_item()
+            if model.should_expand(item, element):
                 row.set_expanded(True)
-            elif row.get_item().element is element:
+            elif item.element is element:
                 if expand:
                     row.set_expanded(True)
                 return n
@@ -395,12 +364,6 @@ def create_search_bar(search_engine: SearchEngine):
     return search_bar
 
 
-def create_shortcut_controller(shortcuts):
-    ctrl = Gtk.ShortcutController.new_for_model(shortcuts)
-    ctrl.set_scope(Gtk.ShortcutScope.LOCAL)
-    return ctrl
-
-
 def create_popup_controller(tree_view, selection, modeling_language):
     menus: dict[str, Gtk.PopoverMenu] = {}
 
@@ -443,14 +406,14 @@ def toplevel_popup_model(modeling_language) -> Gio.Menu:
 
 
 def list_item_factory_setup(
-    _factory, list_item, selection, event_manager, modeling_language
+    _factory, list_item, selection, event_manager, modeling_language, template
 ):
     builder = Gtk.Builder()
     builder.set_current_object(list_item)
     builder.extend_with_template(
         list_item,
         type(list_item).__gtype__,
-        translated_ui_string("gaphor.ui", "treeitem.ui"),
+        template,
         -1,
     )
     row = builder.get_object("row")
@@ -584,7 +547,9 @@ def list_item_drop_drop(
     tree_item = list_item.get_item().get_item()
     dest_element = tree_item.element
     if y < 4:
-        dest_element = dest_element.owner
+        dest_element = owner(dest_element)
+        if dest_element is Root:
+            dest_element = None
 
     with Transaction(event_manager) as tx:
         # This view is concerned with owner relationships.
@@ -611,9 +576,10 @@ def popup_model(element, modeling_language):
     model.append_section(None, part)
 
     part = Gio.Menu.new()
-    part.append_submenu(
-        gettext("New _Diagram"), create_diagram_types_model(modeling_language, element)
-    )
+    diagram_submenu = create_diagram_types_model(modeling_language, element)
+    if diagram_submenu.get_n_items():
+        part.append_submenu(gettext("New _Diagram"), diagram_submenu)
+
     if any(
         isinstance(element, element_type.allowed_owning_elements)
         for element_type in modeling_language.element_types
@@ -622,8 +588,6 @@ def popup_model(element, modeling_language):
             gettext("New _Element"),
             create_element_types_model(modeling_language, element),
         )
-    if isinstance(element, UML.Package):
-        part.append(gettext("New _Package"), "tree-view.create-package")
     model.append_section(None, part)
 
     part = Gio.Menu.new()
@@ -666,13 +630,8 @@ def create_diagram_types_model(modeling_language, element=None):
             )
             part.append_item(menu_item)
 
-    model.append_section(None, part)
-
-    part = Gio.Menu.new()
-    menu_item = Gio.MenuItem.new(gettext("New Generic Diagram"), "win.create-diagram")
-    menu_item.set_attribute_value("target", GLib.Variant.new_string(""))
-    part.append_item(menu_item)
-    model.append_section(None, part)
+    if part.get_n_items():
+        model.append_section(None, part)
 
     return model
 
